'''

Author: Ilia Kurgansky, May 2020

A model of the logic boxes for the black/white balls, as featured in the book "Q is for Quantum" by Terry Rudolph.

Uses a class to represent the mist of configurations and all of its misterious properties.

The current version of the code is not ideal, but fully functional and reasonably well documented. At least for the

latter reason alone this code is now better than Terry's own MATLAB version.

No cats were harmed in the creation of this program.

'''



import numpy as np

import random



class Mist:  # class used to describe the mist containing ball states

    ignored_chars = ["[", "]", " ", "+"]  # these characters are ignored when parsing the string definitions of mists,

    # i.e. brackets are not needed, neither are positive signs.



    def __init__(self, mist_string, configs=None, signs=None):

        """

        Usage examples (all equivalent):

        m1 = Mist("BB,-WW,BW")

        m1 = Mist("[BB,-WW,BW]")

        m1 = Mist("[+BB,-WW,+BW]")

        m1 = Mist(None, ["BB","WW","BW"],["+","-","+"])

        m1 = Mist(None, some_configs, some_signs)



        :param mist_string: the mist can be initialised as a string of the form e.g."[WW, -BW]"

        :param configs: alternatively the mist can be initialised as a list of unsigned configurations, e.g.["WW","BW"]

        :param signs: the unsigned configurations have to be accompanied by a list of signs for each e.g. ["+", "-"]

        """

        if configs is None or signs is None:  # if a list of configurations and signs is missing

            if mist_string is not None:  # make sure there is a string description of configurations

                for char in Mist.ignored_chars:

                    mist_string = mist_string.replace(char, "")  # remove unimportant characters from config string

                mist_list = mist_string.split(",")  # split by comma to create a list of (signed) mist configurations

                self.signs = ["+" for _ in mist_list]  # initialise an all-positive list of signs for every state

                self.configs = []  # initialise an empty list of configs

                for i, configuration in enumerate(mist_list):  # iterate over the signed mist configurations

                    if "-" in configuration:  # if a negative sign is present...

                        if configuration.index("-") != 0:  # ...ensure it is placed at the front of the given config

                            raise ValueError("Why is your minus in a weird place in the mist?")

                        self.signs[i] = "-"   # record the negative sign for this state (positive states keep positive)

                        self.configs.append(configuration[1:].upper())  # record the unsigned configuration separately

                    else:

                        self.configs.append(configuration.upper())  # if config already positive (unsigned), record it

            else:  # if the init function call lacks a list of states/signs and a string description - raise an error.

                raise ValueError("You did not give me any valid information to form a mist.")

        else:  # if a list of states/signs is provided...

            self.configs = configs  # ...assign the list of unsigned states to instance variable...

            self.signs = signs  # ...and the list of signs for each state as a separate instance variable.



        self.length = None  # initialise these auxiliary instance variables, this one is for number of configs in mist.

        self.ballnum = None  # this variable will store number of balls in the mist(i.e. length of each unsigned config)

        self.recalculate_properties()  # auxiliary instance variables calculated and validated, config space simplified.



    def __repr__(self):

        """

        This function provides a representation for the mist instance. A mist can be represented by a string listing

        of signed configurations appropriately punctuated and enclosed in brackets, e.g. "[WW,-BW]"

        :return: returns the string representation of the signed configurations in the mist

        """

        repr_string = "The mist magically clears and you see: ["  # start with an opening bracket and a fitting message

        for i, config in enumerate(self.configs):  # iterate over unsigned configurations in the mist

            if self.signs[i] == "-":  # where the configuration is allocated a negative sign

                repr_string += "-"+config  # prefix with a "-" and then attach the config string

            else:

                repr_string += config  # if the config is positive, omit the sign and just attach the config

            if i + 1 != self.length:  # for all but the last configuration in the list...

                repr_string += ","  # ...attach a comma to the string. (config before closing brackets doesn't need one)

        repr_string += "]"  # finally attach the closing bracket.

        return repr_string  # return string to serve as the representation for the instance of the mist.



    def apply_sign(self, sign):

        """

        Allows "acting" on the given list by a positive or a negative sign.

        Positive sign has no effect on the signs of configurations in the mist.

        Negative sign inverts the attached signs of each of the configurations

        e.g. +[WW,-BW] = [WW,-BW]  (stored as ["+","-"])

            -[WW,-BW] = [-WW,BW]   (stored as ["-","+"])

        :param sign: "+" or "-", the sign acting on the mist

        :return: None. Is an in-place function.

        """

        if sign == "+":  # if acting sign is positive, make no change

            pass

        else:  # otherwise if acting sign is negative

            self.signs = ["-" if sign == "+" else "+" for sign in self.signs]  # flip the signs in the instance list.



    def right_multiply(self, other):

        """

        The function that performs the tensor(distributive) product of two mists and returns a new mist containing the

        resulting list of configurations.

        e.g. [W,B][WW,-WB] = [WWW,-WWB,BWW,-BWB]

        note that "self" is the left mist, "other" is the right mist in this process. Ordering of balls inside each

        configuration is preserved with the left mist containing the first ball etc. Also note that signs are preserved

        for each configuration.

        :param other: the mist that this current mist is being right-multiplied by.

        :return: a new instance of mist containing the result of the product and the allocated config signs

        (with special cases of either of the mists in the product being empty)

        """

        if self.length == 0:  # if the current mist is empty...

            return other  # ...the result of the product is the other mist.

        elif other.length == 0:  # if the other mist is empty...

            return self  # ...then this mist serves as the result (if both empty, an empty mist is returned either way)

        else:  # if neither mist is empty, the product needs to be evaluated properly

            new_configs = []

            new_signs = []

            for i, config1 in enumerate(self.configs):  # iterate over each config in this mist

                for j, config2 in enumerate(other.configs):  # and iterate over each config in the right mist

                    new_configs.append(config1 + config2)  # concatenate the configs (in all combinations)

                    if self.signs[i] == other.signs[j]:  # if both configs positive/negative - then product is positive

                        new_signs.append("+")  # record a positive sign for current config product

                    else:  # if one config positive, other negative (any order), then outcome is negative

                        new_signs.append("-")  # record a negative sign for current config product

            return Mist(None, new_configs, new_signs)  # initialise a new mist not from string, but from lists



    def update_config(self, config_index, ball_index, new_value):

        """

        The function is necessary to change a particular colour of a particular ball in a particular configuration.

        e.g. update_config(1, 0, "W") on a mist of ["WW","BW"] (with signs ["+","-"])

        would result in ["WW","WW"] (with signs ["+","-"])

        :param config_index: the index (starting at 0) of the configuration that needs to be changed

        :param ball_index: the index (starting at 0) of the ball in that configuration

        :param new_value: the new value to change the ball to ("W" or "B" only)

        :return: None. Is an in-place operation on the mist

        """

        try:  # this try-except catches attempts at addressing states/balls outside of the index of this mist.

            if new_value in ["B", "W"]:  # if the new provided colour is W or B (i.e. accepted values)

                config = self.configs[config_index]  # locate the required config (can raise IndexError if invalid)

                new_config = config[:ball_index] + new_value + config[ball_index+1:]  # deconstruct and rebuild with new

                self.configs[config_index] = new_config  # replace the config value with the new rebuilt config

            else:  # if the colour is invalid...

                raise ValueError("Wrong colour given: "+str(new_value))  # ...raise a ValueError to notify of bad colour

        except IndexError:  # if either of the addressing above fails due to index error...

            raise IndexError("Cannot update config. Bad ball index or bad config index.")  # ...raise custom IndexError



    def simplify(self):

        """

        Simplifies redundant/destructive configurations in the mist. Returns the simplified configurations and

        corresponding sign allocations for new configurations.

        Important: Preserves the ratio of different configurations in the group.

        e.g. ["WW","WW","WB"] (with signs ["+","+","-"]) is as simple as it gets, because the ratio of configurations

        is 2:1.

        e.g. ["WW","WW","WW","WW","WB","WB"] (with signs ["+","+","+","+","-","-"]) can be simplified to the example

        mist of configurations above, because the ratio is 4:2, simplified as 2:1

        OR

        ["WW","WB","WB"] (with signs ["+","+","-"]) gets simplified to ["WW"] (with signs ["+"])

        Note the destructive interference between opposite signed configurations in the above example.

        :return: Returns the simplified config/signs lists for replacement of the instance variables.

        """

        config_dict = {}  # A dictionary is used to simplify the repeated states. Dict shape {"config": "signed count"}

        new_configs = []

        new_signs = []

        for i in range(len(self.configs)):  # iterate over all the indices of the current mist configurations

            config = self.configs[i]  # look at each configuration

            sign = self.signs[i]  # and at each corresponding sign

            if sign == "+":  # if the sign of a given configuration is positive

                add = 1  # add to the count of the configurations

            else:

                add = -1  # if negative, then subtract it

            config_dict[config] = config_dict.get(config, 0) + add  # update the dictionary of counts



        counts = np.array(list(config_dict.values()))  # convert to a numpy array to be used by a np function below

        common_ratio = np.gcd.reduce(counts)  # find the common ratio that the configuration counts can be reduced by

        for pair in sorted(config_dict.items(), reverse=True):  # sort dictionary by config string and iterate over

            config = [pair[0]]  # current configuration as a string in list brackets (so it can be added to a list)

            count = pair[1]  # the signed count of this config e.g. three same negative configs have count -3

            simplified_count = abs(count)//common_ratio  # simplify by common ratio. If counts mutually prime, no effect

            new_configs += config*simplified_count  # add the (simplified) number of the same config back into the list

            if count>0:  # if the signed count is positive...

                new_signs += ["+"]*simplified_count  # add the appropriate number of "+" entries to the signs list.

            elif count<0:  # if the count is negative...

                new_signs += ["-"]*simplified_count  # add the number of "-" entries to represent the negative configs.

            else:

                pass  # if the count is 0, do nothing



        return new_configs, new_signs  # return the new config lists and new signs to redefine the mist



    def simplify_incorrect_out_of_use(self):

        """

        INCORRECT: This does NOT preserve the relative ratios of different states.

        i.e. [WW,WW,WW,BB,BB] can NOT be simplified, but this function simplifies it to [WW,BB].

        Need to rework



        Previous:

        Simplifies redundant/destructive configurations in the mist. Returns the simplified configurations and

        corresponding sign allocations for new configurations.

        e.g. ["WW","WW","WB"] (with signs ["+","+","-"]) gets simplified to ["WW","WB"] (with signs ["+","-"])

        or   ["WW","WB","WB"] (with signs ["+","+","-"]) gets simplified to ["WW"] (with signs ["+"])

        Note the destructive interference between opposite signed configurations in the second example.

        :return: Returns the simplified config/signs lists for replacement of the instance variables.

        """

        if len(set(self.configs)) == len(self.configs) and False:  # currently disabled condition.

            # This condition used to check if the configuration is as simple as it can be and if so make no changes.

            # This check is currently disabled because a full calculation (even if redundant) results in a nice sort

            # of the configuration list with a corresponding sort of the sign list. Efficiency gain due to this check

            # is not worth losing the sorting of outcomes in the final presentation.

            return self.configs, self.signs

        else:

            config_dict = {}  # A dictionary is used to simplify the repeated states. Dict of shape: {"config": "sign"}

            new_configs = []

            new_signs = []

            for i in range(len(self.configs)):  # iterate over all the indices of the current mist configurations

                config = self.configs[i]  # look at each configuration

                sign = self.signs[i]  # and at each corresponding sign

                if config in config_dict.keys():  # if current configuration already exists as key in the dictionary...

                    if sign != config_dict[config]:  # ...and current instance has an opposite sign to previous instance

                        config_dict.pop(config)  # ...then destructively interfere and remove current configuration.

                    else:

                        # if a version with this sign already exists in the dictionary, do nothing new.

                        # This configuration already recorded, no duplicates necessary.

                        pass

                else:  # If the current config occurs for the first time...

                    config_dict[config] = sign  # ...store the config string as key and the sign as value

            for pair in sorted(config_dict.items(), reverse=True):  # sort dictionary by config string and iterate over

                new_configs.append(pair[0])  # this way the configs are appended in a sorted order...

                new_signs.append(pair[1])  # ...with the correct corresponding signs appended in the same order.



            return new_configs, new_signs



    def validate_colours(self):

        """

        Upon initialisation of the mist (and further calculations on the mist) check that no erroneous colours are

        present in any of the configurations. If so, raise a ValueError and cease the program.

        :return: None. Intended to crash the program if anything goes wrong.

        """

        for config in self.configs:  # check every config

            for ball in config:  # and every ball in that config

                if ball != "B" and ball != "W":  # if the colour of the ball is not B, nor W

                    raise ValueError("The allowed colours are B and W, I don't know what to do with "+ball)  # error



    def validate_ballnumber(self):

        """

        All ball configurations in the mist must contain the same number of balls. If the number of balls is equal for

        each configuration, then min(number of balls) = max(number of balls) by definition.

        This function checks if this statement is true. If not an exception is raised.

        :return: None. Meant to crash the program with an error if bad data is present.

        """

        config_lengths = [len(config) for config in self.configs]  # generate a list of numbers of balls in each config.

        try:  # the list above is empty for an empty mist, so following conditional check would crash with an error

            if min(config_lengths) != max(config_lengths):  # if there is a discrepancy in the config lengths...

                print(self.configs)  # ...display the configs for reference

                raise ValueError("The mist provided has inconsistent configurations.")  # and raise an error.

        except ValueError:

            pass



    def recalculate_properties(self):

        """

        After the initialisation of the mist or after an in-place calculation (such as a NOT gate etc) some auxiliary

        instance variables need to be (re)calculated, configs need to be simplified and sorted, also a good time to

        check that no erroneous data was introduced to the mist (e.g. bad colours or inconsistent config states)

        :return: None. Either calculates new instance parameters in-place or raises exceptions if problems.

        """

        self.configs, self.signs = self.simplify()  # reassign the configs and signs lists after simplification/sort

        self.length = len(self.configs)  # precalculate the number of configs in the mist to save the effort later

        try:  # try calculating the number of balls in an arbitrary configuration. Set it to 0 if mist is empty

            self.ballnum = len(self.configs[0])

        except IndexError:  # if mist is empty, the above line would raise an index error. Suppress and record 0 balls.

            self.ballnum = 0

        self.validate_colours()  # check colours are valid B or W

        self.validate_ballnumber()  # check ball number in each config is consistent.





def validate_indices(mist, index_lst):

    """

    Before evaluating any gate check that the instructions to address specific indices in the mist are valid.

    :param mist: mist that the gate is being applied to

    :param index_lst: the list of all indices provided to the gate. If any of them are invalid - raise an exception.

    :return: None. Meant to raise an error if the input is invalid.

    """

    for index in index_lst:  # check every index in the list of parameters

        if index < 0 or index >= mist.ballnum:  # if the index is negative or above the number of balls in the mist

            raise IndexError("The index "+str(index)+" is larger than the number of balls in "+str(mist))  # error





def simple_cswap(control, target1, target2):

    """

    A non-misterious version of the cswap gate. Target1 and Target2 swap values if the Control ball is B.

    :param control: "B" or "W"

    :param target1: "B" or "W"

    :param target2: "B" or "W"

    :return: swapped values of the two targets

    """

    if control == "B":

        return simple_swap(target1, target2)

    else:

        return target1, target2





def simple_not(ball):

    """

    A non-misterious version of the not gate. Flips the colour of the ball.

    :param ball: "B" or "W"

    :return: the flipped colour of the target ball

    """

    if ball == "W":

        return "B"

    else:

        return "W"





def simple_cnot(control, target):

    """

    A non-misterious version of the cnot gate. If the control is B, then the target colour is flipped.

    Uses the simple_not gate.

    :param control: "B" or "W"

    :param target: "B" or "W"

    :return: Flipped colour of target if control is B

    """

    if control == "B":

        return simple_not(target)

    else:

        return target





def simple_swap(ball1, ball2):

    """

    A non-misterious version of swap. Swaps the two colours of the two targets and returns

    :param ball1: "B" or "W"

    :param ball2: "B" or "W"

    :return: swapped colours of two balls

    """

    return ball2, ball1





def simple_pete(ball):

    """

    A single-ball version of the PETE box. Returns mists depending on the unsigned ball colour. Is used in the misty

    version of the PETE box.

    :param ball: "B" or "W"

    :return: A mist instance depending on the colour of the input ball.

    """

    if ball == "W":

        return Mist("[W, B]")

    else:

        return Mist("[W, -B]")





def OBSERVE(mist, target_list=None, clairvoyance=0):

    """

    Usage examples:

    m2 = OBSERVE(m1)

    m2 = OBSERVE(m1, clairvoyance=2)

    m2 = OBSERVE(m1, [0,2,4])

    m2 = OBSERVE(m1, [0,2,4], 9001)



    All misty functions return a mist.

    Observing a mist is (sometimes unintended) transformative action. Just like the boxes, an observation changes

    the mist, and, unlike many of the boxes, this change is irreversible.

    The function emulates the outcome of an observation, not its process, because we know very little about the process.

    A mist is taken as a parameter, with an optional list of specific ball indices to be observed.

    By default all available balls are observed.

    Firstly, all possible outcomes of the observation are collated, alongside the config states that would match that

    observation.

    Then, a weighted random draw is made out of the list of outcomes, based on the square of the frequency addition rule

    e.g. if the possible observation "W_B_" (ball 0 and ball 2 are observed, the other two are left untouched) has

    3 possible states that would result in this observation:

    e.g. [WWBW, WBBW, -WWBB] then the relative weight of this observation is 3^2 = 9.

    If there exists only one other option of "B_B_", which is supported by e.g. 2 states, then the weight of observing

    "B_B_" is 2^2 =4. The respective probabilities of observing "W_B_" or "B_B_", are, as defined in part 2 of the book,

    given as 9/(4+9) and 4/(4+9)

    The function operates with more information that would normally be available to an observer, so it allows for the

    parameter of clairvoyance that lets the observer receive information from beyond the naturalistic veil.

    The function returns the mist after the randomly drawn observation "pins" the states of the mist to conform to it.

    :param mist: the mist of configurations to be selectively observed

    :param target_list: (optional) a list of indices for balls to be observed

    :param clairvoyance: (optional) 0 by default, otherwise 1, 2 or over 9000. Higher levels reveal more information.

    :return: returns the "collapsed" mist, i.e. a mist of configurations that allow the observed outcome.

    """

    if target_list is None:  # if the optional parameter of ball indices is missing

        target_list = list(range(mist.ballnum))  # create a list that covers every index

    validate_indices(mist, target_list)  # check the supplied list of indices for validity

    possible_combos = {}  # stringy combo : [matching configs], will store the possible outcomes of the observation

    for i, config in enumerate(mist.configs):  # iterate over all of the configurations in the mist

        # then group the configurations by the observable values at desired indices.

        new_combo = list("_"*mist.ballnum)  # start with a blank description of the observation

        for target_index in target_list:

            new_combo[target_index] = config[target_index]

        new_combo = "".join(new_combo)  # this is the description of the observation that the current config would give.

        possible_combos[new_combo] = possible_combos.get(new_combo, []) + [(config, mist.signs[i])]  # group configs



    prob_denominator = sum(len(configlst)**2 for configlst in possible_combos.values())  # used for probability

    combos = sorted(possible_combos.keys(), reverse=True)  # a list of unique combos that could come from an observation

    combo_weights = [len(possible_combos[combo])**2/prob_denominator for combo in combos]  # list of probabilities



    observation = random.choices(combos, combo_weights)[0]  # random weighted choice from the available combinations



    new_configs = [config[0] for config in possible_combos[observation]]  # these configs allow for the chosen obs.

    new_signs = [config[1] for config in possible_combos[observation]]  # matching signs for the configs

    collapsed_mist = Mist(None, new_configs, new_signs)  # mist that accommodates for the outcome of the observation



    print("You observe:", observation, end=" ")  # normally available information

    if clairvoyance >= 1:

        prob_obs = combo_weights[combos.index(observation)]

        print("with probability", round(prob_obs,3))  # additional information for weak quantum psychics

    else:

        print()

    if clairvoyance >= 2:

        for i, combo in enumerate(combos):

            if combo != observation:

                print("Another possible observation is", combo, end=" ")

                print("with probability", round(combo_weights[i], 3))  # stronger psychics know of other outcomes.

    if clairvoyance > 9000:

        print("Your collapsed mist is", collapsed_mist)  # strongest psychics can perceive superpositions of states.



    return collapsed_mist





def COMBINE(mist_list):

    """

    Usage examples:

    total = COMBINE([m1,m2,m3])



    Combines a list of mists into a single mist by using a tensor product (using iterative right-multiplication).

    e.g. COMBINE([m1,m2,m3]) is equivalent to doing m1.right_multiply(m2).right_multiply(m3)

    :param mist_list: A list of mists/mist references, e.g. [m1,m2,m3]

    :return: a new instance of a mist with the product of configurations of initial mists

    """

    if len(mist_list) == 0:  # if the list of mists is empty, complain about it

        raise IndexError("You gave me no mists to combine. What am I to do?")

    subtotal = mist_list[0]  # start with the leftmost mist in the list

    for i in range(1,len(mist_list)):  # iterate over the remaining mists

        other = mist_list[i]

        subtotal = subtotal.right_multiply(other)  # perform the right multiplication by each remaining mist

    return subtotal  # return the final outcome





def NOT(mist, target_index):

    """

    Usage example:

    m2 = NOT(m1, 3)



    All misty gates return mists

    A misty version of the not gate. Needs to know which ball in the mist to apply the not transformation to in every

    configuration. Does not initialise a new mist, but returns reference to the same amended mist.

    :param mist: Mist of configurations to apply the gate to.

    :param target_index: Index of the ball in each configuration of the mist to flip colour.

    :return: reference to the (in-place) amended mist instance

    """

    validate_indices(mist, [target_index])  # first check that the indices are valid

    for i, config in enumerate(mist.configs):  # iterate over all configurations in this mist

        new_value = simple_not(config[target_index])  # flip the colour of the corresponding ball in each config

        mist.update_config(i, target_index, new_value)  # assign this new value to the specific ball in each config

    mist.recalculate_properties()  # refresh the mist, which involves sorting of the config list for prettiness.

    return mist  # return reference to the same amended mist (not a new mist instance)





def SWAP(mist, target1_index, target2_index):

    """

    Usage example:

    m2 = SWAP(m1, 0, 4)



    All misty gates return mists.

    A misty version of the swap gate. Needs to know which two balls in the mist to apply the swap transformation to in

    every configuration. Does not initialise a new mist, but returns reference to the same amended mist.

    :param mist: which mist of configurations to apply the box to

    :param target1_index: index of swap target1 in each of the configurations

    :param target2_index: index of swap target2 in each of the configurations

    :return:

    """

    validate_indices(mist, [target1_index, target2_index])  # check the indices are valid first

    for i, config in enumerate(mist.configs):  # iterate over the configurations in the mist

        new_ball1, new_ball2 = simple_swap(config[target1_index], config[target2_index])  # get new values for config

        mist.update_config(i, target1_index, new_ball1)  # update the mist with the changes of the first ball

        mist.update_config(i, target2_index, new_ball2)  # update the mist with the changes of the second ball

    mist.recalculate_properties()  # recalculate and sort the mist for presentability

    return mist  # return reference to the same amended mist





def CNOT(mist, control_index, target_index):

    """

    Usage example:

    m2 = CNOT(m1, 0, 3)



    All misty gates return mists

    A misty version of the cnot gate. Needs to know which ball in the mist to apply the not transformation to in every

    configuration, and which ball will serve as control in every configuration. If control is B - flip the target.

    Does not initialise a new mist, but returns reference to the same amended mist.

    :param mist: mist to apply the gate to.

    :param control_index: index of the control ball in every configuration in the mist

    :param target_index: index of the target ball in every configuration in the mist

    :return: reference to the same amended mist

    """

    validate_indices(mist, [control_index, target_index])  # check that all indices are valid first

    for i, config in enumerate(mist.configs):  # iterate over all configurations in the mist

        if config[control_index] == "B":  # if the control ball in each configuration is B

            new_value = simple_not(config[target_index])  # perform the not box as usual

            mist.update_config(i, target_index, new_value)  # and update the mist (do nothing otherwise)

    mist.recalculate_properties()  # recalculate/sort the mist configurations

    return mist  # return reference to the amended mist





def CCNOT(mist, control1_index, control2_index, target_index):

    """

    Usage example:

    m2 = CCNOT(m1, 0, 2, 4)



    All misty gates return mists

    A misty version of the ccnot gate, as featured in the book, part 1. Changes the colour of the target ball iff both control

    balls are B.

    Made from a stack of CNOT-CSWAP-CNOT gates, with the CNOT gates having flipped control-target inputs to the cswap.

    Refer to the book for illustration.

    Does not initialise a new mist, but returns reference to the same amended mist.

    :param mist: mist to operate the box on

    :param control1_index: index of first control in every configuration

    :param control2_index: index of second control in every configuration

    :param target_index: index of the target ball in every configuration

    :return: reference to the same amended mist.

    """

    validate_indices(mist, [control1_index, control2_index, target_index])  # check the indices are valid

    mist = CNOT(mist, target_index, control2_index)  # perform a CNOT (flipped control and target)

    mist = CSWAP(mist, control1_index, control2_index, target_index)  # perform a cswap

    mist = CNOT(mist, target_index, control2_index) # perform another flipped CNOT.

    mist.recalculate_properties()  # recalculate and sort the mist

    return mist  # return reference to the same amended mist





def CSWAP(mist, control_index, target1_index, target2_index):

    """

    Usage example:

    m2 = CSWAP(m1, 3, 0, 2)



    All misty gates return mists

    A misty version of the cswap gate. Swaps the values of the two target balls in each configuration, if the control

    ball in the corresponding configuration is B.

    Does not initialise a new mist, but returns reference to the same amended mist.

    :param mist: mist to operate the box on

    :param control_index: index of the control ball in each configuration

    :param target1_index: index of the first target ball in each configuration

    :param target2_index: index of the second target ball in each configuration

    :return: reference to the same amended mist

    """

    validate_indices(mist, [control_index, target1_index, target2_index])  # check the indices are valid first

    for i, config in enumerate(mist.configs):  # iterate over all configurations in the mist

        if config[control_index] == "B":  # if the corresponding control ball is B

            new_ball1, new_ball2 = simple_swap(config[target1_index], config[target2_index])  # perform a swap as usual

            mist.update_config(i, target1_index, new_ball1)

            mist.update_config(i, target2_index, new_ball2)

    mist.recalculate_properties()  # recalculate and sort the mist

    return mist  # return reference to the same amended mist





def PETE(mist, target_index):

    """

    Usage example:

    m2 = PETE(m1, 0)



    All misty boxes return mists.

    The crazy PETE box. Produces a two-config mist in place of the ball in each configuration. This function "expands"

    the misty configurations, which leads to an increase in configurations overall. Returns a new instance of a mist

    with a simplified list of the configurations, so no redundant/pre-interference configurations are seen.



    Works by splitting a configuration into three pieces: pre-target "left" section, the target peted ball and the

    post-target "right" section. e.g. a configuration "WWBW" with target index 2 is split into "WW", "B", "W".



    The target ball is then passed through a Pete box, resulting in a mist.

    Each of the "left" and the "right" configs are used to define a single-config mist with their matching signs.

    The configurations are then reconstructed by finding the product of the "left-mist","target-mist" and "right-mist".



    These resulting configurations (with matching signs) are added onto an overall list of configurations, which is then

    used to initialise a new mist instance.



    :param mist: mist to operate the box on

    :param target_index: the index of the target ball in each configuration

    :return: a new mist instance with a new calculated configuration list with all brackets expanded and simplified

    """

    new_config = []

    new_signs = []

    for i, config in enumerate(mist.configs):  # iterate over all configurations in the mist

        config_left_mist = Mist(None, [config[:target_index]], [mist.signs[i]])  # make pre-target "left" mist



        peted_target_mist = simple_pete(config[target_index])  # apply PETE to the unsigned target ball

        peted_target_mist.apply_sign(mist.signs[i])  # apply the matching configuration sign to the peted-target mist



        config_right_mist = Mist(None, [config[target_index+1:]], [mist.signs[i]])  # make a post-target "right" mist



        recombined_mist = COMBINE([config_left_mist, peted_target_mist, config_right_mist])  # "multiply" the mists



        new_config += recombined_mist.configs  # add the resulting configurations iteratively

        new_signs += recombined_mist.signs  # add the matching signs for each configuration iteratively



    return Mist(None, new_config, new_signs)  # generate a new mist from the new configurations and return reference





def ARCHIMEDES(mist, correct_locations = None):

    """

    Usage examples:

    outcome = ARCHIMEDES(m1)

    outcome = ARCHIMEDES(m1, ["WWW", "WBB"])



    All misty boxes return mists.

    The default 4-ball Archimedes vault computer, as described in the book. Validates the positions of golden bars.

    If 3-ball position matches one of the correct locations, then it applies a NOT to the fourth ball.

    Default correct_locations are the same ones as the examples in the book. Alternative correct_locations can be

    provided.

    Has no validation. Incorrect numbers of input balls or invalid correct locations may destroy this box (and take the

    program with it). Your loss, however, because you are now stuck in a bank vault with no way of finding the real

    golden bars.

    :param mist: The 4-ball mist passing through Archimedes. Three balls for control (location), one for target.

    :param correct_locations: Optional parameter. List of strings of real gold locations. Case-sensitive, so be careful.

    :return: Returns reference to the mist of the four balls after all of the in-place manipulations.

    """

    if correct_locations is None:  # if an alternative list of golden bar locations is not provided, use the example.

        correct_locations = ["WWB", "WBW", "BWB", "BBB"]



    for i, config in enumerate(mist.configs):  # iterate over all configurations in the mist

        if config[:(mist.ballnum-1)] in correct_locations:  # if the first three balls match a correct location

            new_value = simple_not(config[mist.ballnum-1])  # then apply a "not" to the fourth ball

            mist.update_config(i, mist.ballnum-1, new_value)

    mist.recalculate_properties()  # recalculate/sort/simplify the mist

    return mist  # return reference to the amended mist.





def NARCHIMEDES(mist):

    """

    Usage example:

    outcome = NARCHIMEDES(m1)



    An n-hole variation of the Archimedes box. Magically adjusts to the given mist to have enough inputs and outputs.

    e.g. a 4-ball mist (3 control and 1 target) will play the 8 golden bar = 4 fake + 4 real game.

    e.g. a 6-ball mist (5 control and 1 target) can map 2^5 = 32 golden bar positions, of which half (2^4) are real.

    Correct locations of real golden bar are randomised to fit the size of the scenario, always having half of the bars

    being real.

    Uses the default Archimedes function and provides the customised correct locations list for it.

    :param mist: the post-pete mist of input balls (i.e. the entire permutation space of the n-ball configuration)

    :return: Returns a reference to the same amended mist.

    """

    real_gold_num = 2**((mist.ballnum-1)-1)  # half of all bars mappable by this mist could be real gold

    possible_locations = list(set([config[:(mist.ballnum-1)] for config in mist.configs]))  # locations from the mist

    random_correct_locations = random.sample(possible_locations, real_gold_num)  # sample half of those for real gold

    print("Psst... the real golden bars are at:", random_correct_locations)  # let the user know (secretly)

    return ARCHIMEDES(mist, random_correct_locations)  # run the ARCHIMEDES logic





def Bank_robbery():

    """

    Calculation of the bank vault scenario, exactly as described in the book.

    Have a go at understanding what each part of the code represents from the narrative.

    """

    input_mist = Mist("WWWB")

    pre_Archimedes = input_mist



    for i in range(input_mist.ballnum):

        pre_Archimedes = PETE(pre_Archimedes, i)

    print(pre_Archimedes)



    post_Archimedes = ARCHIMEDES(pre_Archimedes)

    print(post_Archimedes)



    outcome = post_Archimedes

    for i in range(outcome.ballnum-1):

        outcome = PETE(outcome, i)

    print(outcome)



# ALL INDEXING BEGINS FROM 0





